import argparse
import csv
import json
import logging
import sys

import rapticoressvc.svcc_helper
from rapticoressvc import helpers
from rapticoressvc.vector_calculator_helpers import vector_calculate_utility, vector_calculate_exposure, \
    vector_calculate_exploitability, vector_calculate_impact

combined_results = []


class SetEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)


def xstr(s):
    if s == "None":
        return None
    else:
        return str(s)


def xstr_asset_criticality(s):
    if s == "None":
        return "high"
    else:
        return str(s)


def ssvc_recommendations(asset, vul_details, public_status, environment, asset_type, asset_criticality):
    logging.debug('Initializing......')
    engine_check = helpers.initialize()
    if not engine_check:
        logging.error('DB engine could not be Initialized')
        sys.exit(1)
    logging.debug('DB engine Initialized')
    query = {}
    description = None
    severity_list = ["critical", "high", "medium", "low"]
    severity_priority = ["critical", "high"]
    cvss_vector = None
    nvd_data_local = None
    score = None
    exploit_status = None
    recommendation = None
    if vul_details in severity_list:
        score = vul_details
        if vul_details in severity_priority:
            exploit_status = "PoC"
        else:
            exploit_status = "None"
    else:
        if type(vul_details) is not list:
            vul_details = [vul_details]
        exploit_data = []
        for vul_detail in vul_details:
            try:
                get_data = helpers.input_cve_get_nvd_data(vul_detail)
                if get_data:
                    cvss_vector, score, nvd_data_local = get_data[1], get_data[2], json.loads(get_data[3])
                    if cvss_vector:
                        exploit_status = vector_calculate_exploitability(vul_detail, cvss_vector)
                    description = nvd_data_local["cve"]["description"]["description_data"][0]["value"]
                    exploit_data.append(dict(cvss_vector=cvss_vector, score=score, exploit_status=exploit_status,
                                             description=description))
            except Exception as e:
                logging.exception(e)
                # todo handle this
        max_exploit = max(exploit_data, key=lambda x: x['score'])

        exploit_status = max_exploit.get('exploit_status')
        score = max_exploit.get('score')
        cvss_vector = max_exploit.get('cvss_vector')
        description = max_exploit.get('description')

    query["Exploitation"] = exploit_status
    query["Exposure"] = vector_calculate_exposure(score)
    query["Utility"] = vector_calculate_utility(exploit_status, cvss_vector, public_status, score)
    query["Impact"] = vector_calculate_impact(environment, asset_type, asset_criticality)
    recommendation = rapticoressvc.svcc_helper.calculate_recommendation(query)
    if recommendation:
        recommendation = list(recommendation.keys())[0]
    else:
        recommendation = "review"

    results = dict(asset=asset, description=description, cve=vul_details, vulnerability_score=score,
                   cvss_vector=cvss_vector,
                   asset_type=asset_type, environment=environment,
                   public_status=public_status, asset_criticality=asset_criticality, ssvc_rec=recommendation)

    logging.info(results)
    combined_results.append(results)
    return results


def main():
    logging.getLogger()
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--single', help="Parameter based entry", action='store_true')
    group.add_argument('--datafile', help="csv file upload - use --file option", action='store_true')

    parser.add_argument(
        "-id",
        "--asset_id",
        help="Asset Identifier",
        default=None,
        type=str,
    )

    parser.add_argument(
        "-cn",
        "--cve_number",
        help="CVE numbers for the vulnerability separated by '|'",
        default=None,
        type=str,
    )

    parser.add_argument(
        "-p",
        "--public_status",
        help="Public Status allowed values. Choices: public, public_restricted, private",
        default="None",
        choices=["public", "public_restricted", "private", "None"],
        type=str,
    )
    parser.add_argument(
        "-vs",
        "--vul_severity",
        help="Vulnerability Severity",
        default="None",
        choices=["critical", "high", "medium", "low"],
        type=str,
    )

    parser.add_argument(
        "-e",
        "--environment",
        help="Environment for the asset. Choices: production, non_production, None",
        choices=["production", "non_production", "None"],
        default="None",
        type=str,
    )

    parser.add_argument(
        "-a",
        "--assetType",
        help="Asset Type allowed values. Choices: DB, Compute, Storage, None",
        choices=["db", "compute", "storage", "None", "network"],
        default="None",
        type=str,
    )

    parser.add_argument(
        "-s",
        "--criticality",
        help="Business Criticality of an asset. Choices: critical, high, medium, low",
        choices=["critical", "high", "medium", "low"],
        type=str,
        default="high"
    )

    parser.add_argument(
        "--file",
        help="Provide a vulnerability/host via stdin (e.g. through piping) or --file",
        type=argparse.FileType("r", encoding="utf-8-sig"),
        default=sys.stdin,
    )

    parser.add_argument(
        "-v",
        "--verbose",
        help="Increase output verbosity",
        action="store_true",
        default=False,
    )

    log_level = logging.INFO
    log_format = "%(message)s"

    args = parser.parse_args()

    if args.verbose:
        log_level = logging.DEBUG

    logging.basicConfig(level=log_level, stream=sys.stderr, format=log_format)

    if args.datafile:
        logging.debug('Processing datafile')
        data_file = args.file
        reader = csv.DictReader(data_file)
        for row in reader:
            asset_id = str(row['asset_id']).strip()
            cve_number = str(row['cve_number']).strip()
            cve_number = xstr(cve_number)
            cve_list = None
            if cve_number:
                cve_list = cve_number.split('|')
                cve_list = [cve.strip() for cve in cve_list]

            vul_severity = str(row['vul_severity']).strip()
            environment = str(row['environment']).strip()
            public_status = str(row['public_status']).strip()
            asset_type = str(row['assetType']).strip()
            asset_criticality = str(row['assetCriticality']).strip()
            asset_criticality = xstr_asset_criticality(asset_criticality)

            if cve_list:
                ssvc_recommendations(asset_id, cve_list, public_status, environment, asset_type, asset_criticality)
            elif vul_severity:
                ssvc_recommendations(asset_id, vul_severity, public_status, environment, asset_type, asset_criticality)
    elif args.single:
        logging.debug('Processing single parameter based entry')
        cve_number = str(args.cve_number)
        cve_number = xstr(cve_number)
        vul_severity = str(args.vul_severity)
        if cve_number:
            asset_id = str(args.asset_id)
            environment = str(args.environment)
            public_status = str(args.public_status)
            asset_type = str(args.assetType)
            asset_criticality = str(args.criticality)

            ssvc_recommendations(asset_id, cve_number, public_status, environment, asset_type, asset_criticality)
        elif vul_severity:
            asset_id = str(args.asset_id).rstrip()
            environment = str(args.environment)
            public_status = str(args.public_status)
            asset_type = str(args.assetType)
            asset_criticality = str(args.criticality)
            ssvc_recommendations(asset_id, vul_severity, public_status, environment, asset_type, asset_criticality)
    else:
        parser.print_help()
        exit(1)

    logging.info('Writing results to excel file')
    helpers.excel_writer(combined_results)
    logging.info('Results written to excel file ssvc_recommendations.xlsx')

if __name__ == "__main__":
    main()



import datetime
import json
import logging
import os
import sqlite3
from pathlib import Path
from urllib.request import pathname2url
import requests
import io
import zipfile
from nested_lookup import nested_lookup
import pandas as pd


def execute_nvd_tbl_count():
    db_connection, conn = get_db_conn()[0], get_db_conn()[1]
    data = db_connection.execute(
        "SELECT COUNT(1) FROM nvd")
    data = data.fetchone()
    if data:
        return data[0]
    else:
        return None


def execute_threat_intel_tbl_count():
    db_connection, conn = get_db_conn()[0], get_db_conn()[1]
    data = db_connection.execute(
        "SELECT COUNT(1) FROM threatIntel")
    data = data.fetchall()
    if data:
        return data[0]
    else:
        return None


def execute_nvd_tbl_query(query_data):
    db_connection, conn = get_db_conn()[0], get_db_conn()[1]
    data = db_connection.execute(
        "SELECT * FROM nvd WHERE id=?", [query_data]
    )
    data = data.fetchall()
    if data:
        return data[0]
    else:
        return None


def get_db_conn():
    try:
        connection = sqlite3.connect('threatdb.db')
        cursor = connection.cursor()
        if connection and cursor:
            return cursor, connection
        else:
            return None
    except Exception as e:
        logging.exception(e)
    return None


def download_extract_zip(url):
    response = requests.get(url)
    with zipfile.ZipFile(io.BytesIO(response.content)) as zip_file:
        for zip_info in zip_file.infolist():
            with zip_file.open(zip_info) as the_file:
                file = json.loads(the_file.read())
                return file


def initialize_nvdb_tbl():
    try:
        connection = sqlite3.connect('threatdb.db')
        cursor = connection.cursor()
        cursor.execute(
            "CREATE TABLE IF NOT EXISTS nvd (id varchar (25) NOT NULL PRIMARY KEY,cve_vector varchar(50), "
            "cve_score varchar(50),data json)")

        return True
    except Exception as e:
        logging.exception(e)
    return None


def load_nvd_tbl(nvd_data):
    logging.debug("Loading NVD Table")
    db_connection = None
    conn = None
    db = get_db_conn()
    if db:
        db_connection, conn = db[0], db[1]
    for data in nvd_data["CVE_Items"]:
        cve_id = data["cve"]['CVE_data_meta']['ID']
        impact = data['impact']
        if nested_lookup('vectorString', impact):
            cve_vector = nested_lookup('vectorString', impact)[0]
        else:
            cve_vector = None
        if nested_lookup('baseScore', impact):
            cve_score = nested_lookup('baseScore', impact)[0]
        else:
            cve_score = None
        db_connection.execute("insert into nvd values (?, ?, ?, ?)", [cve_id, cve_vector, cve_score, json.dumps(data)])
    conn.commit()
    conn.close()


def get_nvd_data():
    """
    check first if the data is already present.
    Add a years in the years array to grab additional data from NVD for that year. Reduced

    """
    check = execute_nvd_tbl_count()
    if check:
        return True
    else:
        years = ["2023", "2022", "2021", "2020", "2019", "2018"]
        for year in years:
            get_and_load_nvd_file(year)


def get_and_load_nvd_file(year):
    zip_url = f'https://nvd.nist.gov/feeds/json/cve/1.1/nvdcve-1.1-{year}.json.zip'
    logging.debug(f"Downloading artifact {zip_url}")
    nvd_data = download_extract_zip(zip_url)
    load_nvd_tbl(nvd_data)


def initialize_nvdb():
    check = initialize_nvdb_tbl()
    if check:
        get_nvd_data()
        return True


def initialize_db():
    try:
        connection = sqlite3.connect('threatdb.db')
        cursor = connection.cursor()
        cursor.execute("CREATE TABLE IF NOT EXISTS threatIntel (id varchar (25) NOT NULL PRIMARY KEY, data json)")
        return True
    except Exception as e:
        logging.exception(e)
    return None


def get_cisa_kevc():
    # todo load url from config.yml
    url = "https://www.cisa.gov/sites/default/files/feeds/known_exploited_vulnerabilities.json"
    response = requests.get(url, stream=True)
    db_connection = None
    conn = None
    threat_data = json.loads(response.content)
    db_foo = get_db_conn()
    if db_foo:
        db_connection, conn = db_foo[0], db_foo[1]
    for data in threat_data["vulnerabilities"]:
        db_connection.execute("insert into threatIntel values (?, ?)", [data['cveID'], json.dumps(data)])
        conn.commit()
    conn.close()


def validate_db_time():
    try:
        if os.path.exists('./threatdb.db'):
            c_time = os.path.getctime('./threatdb.db')
            dt_c = datetime.datetime.fromtimestamp(c_time)
            today = datetime.datetime.now()
            delta = today - dt_c
            days_from_creation = delta.days
            return days_from_creation > 0
    except Exception as e:
        logging.exception(e)
    return False


def initialize():
    try:
        dburi = 'file:{}?mode=rw'.format(pathname2url('threatdb.db'))
        db_validation = validate_db_time()
        path = Path('./threatdb.db')
        if path and db_validation:
            os.remove("./threatdb.db")
            logging.info('threatdb cleaned up')
        check = sqlite3.connect(dburi, uri=True)
        if check:
            return True
    except sqlite3.OperationalError:
        try:
            db_connection = initialize_db()
            if db_connection:
                get_cisa_kevc()
                check = initialize_nvdb()
                # todo add check whether the db was loaded.
                return check
        except Exception as e:
            logging.exception(e)


def execute_db(query_data):
    db_connection, conn = get_db_conn()[0], get_db_conn()[1]
    data = db_connection.execute(
        "SELECT * FROM threatIntel WHERE id=?", [query_data]
    )
    data = data.fetchall()
    if data:
        return data
    else:
        return None


def process_cvss_score(value):
    c_score = value.split("/")
    result = {}
    for val in c_score:
        key = val.split(":")[0]
        s_value = val.split(":")[1]
        result[key] = s_value
    return result


def nvd_parser(cvss_vector):
    score_dict = process_cvss_score(cvss_vector)
    if score_dict.get("E"):
        if score_dict.get("E") == "U":
            return "None"
        if score_dict.get("E") == "P":
            return "PoC"
        if score_dict.get("E") == "F" or score_dict.get("E") == "H":
            return "active"
    else:
        return "None"


def select_all_tasks(conn):
    cur = conn.cursor()
    cur.execute("SELECT * FROM threatIntel")
    rows = cur.fetchall()
    return rows


def input_cve_get_nvd_data(cve_number):
    nvd_data = execute_nvd_tbl_query(cve_number)
    return nvd_data


def excel_writer(data):
    try:
        # Create a Pandas Excel writer using XlsxWriter as the engine.
        writer = pd.ExcelWriter('ssvc_recommendations.xlsx', engine='xlsxwriter')
        df1 = pd.DataFrame(data)
        # Write each dataframe to a different worksheet.
        df1.to_excel(writer, sheet_name='ec2')

        # Close the Pandas Excel writer and output the Excel file.
        writer.close()
    except Exception as e:
        logging.exception(e)



"""
This library is an implementation of Stakeholder Specific Vulnerability Categorization(SVCC)
"""
__version__ = "0.0.13"

from .helpers import initialize_db, get_db_conn, get_cisa_kevc
from .ssvc_ore import ssvc_recommendations



from rapticoressvc.helpers import execute_db, process_cvss_score, nvd_parser
from rapticoressvc.svcc_helper import calculate_impact, calculate_utility


def vector_calculate_exploitability(cve_number, cvss_vector):
    """
    EXPLOITATION = ["none", "PoC", "active"]
    active > PoC > none
    None: There is no evidence of active exploitation and no public proof of concept (PoC)
    of how to exploit the vulnerability.

    Public PoC: One of the following is true: (1) Typical public PoC exists in sources such as
    Metasploit or websites like ExploitDB; or (2) the vulnerability has a well-known
    method of exploitation. Some examples of condition (2) are open-source web
    proxies that serve as the PoC code for how to exploit any vulnerability in the vein
    of improper validation of Transport Layer Security (TLS) certificates, and
    Wireshark serving as a PoC for packet replay attacks on ethernet or Wi-Fi
    networks.

    Active:Shared, observable, and reliable evidence that cyber threat actors

    1 - Check CISA KEVC - check whether a CVE has an active exploit
        if yes =>> Exploitcation = active
    2 - Check NVBlib - Check whether schema has addition information
        if not CISA KEVC ==>
        Check CVSS for Exploit Code Maturity(E)
        Exploit Code Maturity (E)
        ==> Empty: None
            Unprove: E:U
            Proof-of-concept: E:P -- Explication = PoC
            Functional: E:F -- Explication = active
            High: E:H -- Explication = active
    """
    check = execute_db(cve_number)
    if check:
        exploit_status = "active"
        return exploit_status
    else:
        exploit_status = nvd_parser(cvss_vector)
        return exploit_status


def vector_calculate_exposure(score):
    """
    EXPOSURE = ["unlikely", "probable", "unavoidable"]
    unavoidable > controlled > limited
    Partial:
    One of the following is true: The exploit gives the threat actor limited control over,
    or information exposure about, the behavior of the software that contains the
    vulnerability; or the exploit gives the threat actor a low stochastic opportunity for
    total control. In this context, “low” means that the attacker cannot reasonably
    make enough attempts to overcome obstacles, either physical or security-based,
    to achieve total control. A denial-of-service attack is a form of limited control over
    the behavior of the vulnerable component.
    Partial => limited, controlled
    C:L or I:L or A:H and PR:H
    CVSS Score 0/low/medium --> 6.9 => unlikely exposure
    CVSS Score 7 --> 8.9/high == probable exposure
    CVSS Score 9 --> 10/critical == unavoidable exposure
    """
    severity_list = ["critical", "high", "medium", "low"]
    severity_priority = ["critical", "high"]
    severity_defer = ["medium", "low"]
    if score in severity_list:
        if score == "critical":
            return "unavoidable"
        elif score == "high":
            return "probable"
        elif score in severity_defer:
            return "unlikely"
    elif score:
        score = float(score)
        if score >= 9.0:
            return "unavoidable"
        elif 7.0 <= score <= 8.9:
            return "probable"
        elif score <= 6.9:
            return "unlikely"


def vector_calculate_utility(exploit, cvss_vector, public_status, score):
    """
    Effort required to use the vulnerability
    Utility = ["laborious", "complex", "effortless"]

    Effortless - point and click, little to no User Interaction - skills required minimum - easily discoverable
    Complex - Requires skills and targeting, user interaction required even if publicly accessible
    Laborious - Multiples factors, inaccessible, requires skills, user interaction etc.

    effortless > complex > laborious

    Super effective =>
    Exploit + Network Exploit(AV:N) + Publicly Access + User Interaction is NONE(UI:N)
    Exploit + Network Exploit(AV:N) + Publicly Restricted Access + User Interaction is NONE (UI:N)

    Efficient =>
    Exploit + Network Exploit(AV:N) + Publicly Access + User Interaction is REQUIRED (UI:R)
    Exploit + Network Exploit(AV:N) + Public Restricted Access + User Interaction is REQUIRED (UI:R)
    Network Exploit(AV:N) + Publicly Accessible + User Interaction is NONE(UI:N)
    Network Exploit(AV:N) + Public Restricted Accessible + User Interaction is NONE (UI:N)

    Laborious =>
    Network Exploit(AV:N) + PRIVATE + User Interaction is (UI:R)
    Network Exploit(AV:N) + PRIVATE + User Interaction is (UI:N)
    Network Exploit(AV:L) + PRIVATE + User Interaction is (UI:N)
    Network Exploit(AV:L) + PRIVATE + User Interaction is (UI:R)

    https://mitre-attack.github.io/attack-navigator/
    (1) Reconnaissance:
    the vulnerable component is not searchable or enumerable on the network,

    (2) Weaponization may require human direction for each target,

    (3) delivery may require channels that widely deployed network security configurations block, and
    Future Implementation

    (4) exploitation may be frustrated by adequate exploit-prevention techniques
    enabled by default (address space layout randomization [ASLR] is an example of
    an exploit-prevention tool).

    Steps 1-4 of the kill chain—reconnaissance, weaponization, delivery, and
    exploitation — cannot be reliably automated for this vulnerability.1 Examples for
    explanations of why each step may not be reliably automatable include:

    Steps 1-4 of the kill chain can be reliably automated. If the vulnerability
    allows unauthenticated remote code execution (RCE) or command injection, the
    response is likely yes.


    """
    severity_list = ["critical", "high", "medium", "low"]
    severity_priority = ["critical", "high"]
    attack_vector, user_interaction = None, None
    if cvss_vector:
        vector_dict = process_cvss_score(cvss_vector)
        query = {"exploit": exploit, "attack_vector": vector_dict.get("AV"), "user_interaction": vector_dict.get("UI"),
                 "public_status": public_status}
        utility = calculate_utility(query)
        return utility
    else:
        if score == "critical":
            attack_vector, user_interaction = "N", "N"
        elif score == "high":
            attack_vector, user_interaction = "N", "R"
        else:
            attack_vector, user_interaction = "L", "R"
        query = {"exploit": exploit, "attack_vector": attack_vector, "user_interaction": user_interaction,
                 "public_status": public_status}
        utility = calculate_utility(query)
        return utility


def vector_calculate_impact(environment, asset_type, asset_criticality):
    """
    Very High, High = Environment Production - Asset Type DB, Compute, Storage, Critical, High
    High, medium, low = Environment Production - Asset Type Everything else, high, medium, low
    High = Environment Non-production - Asset Type DB, Compute, Storage, critical, high, medium, low
    High, medium, low = Environment Non-production - Asset Type Everything else, High, medium, low

    IMPACT = ["low", "medium", "high", "very high"]

    Minimal: Neither support nor essential apply. The vulnerable component may be used within the
    entities, but it is not used as a mission-essential component, nor does it provide
    impactful support to mission-essential functions.
    Support The vulnerable component only supports MEFs for two or more entities.
    Asset Criticality = ["low", "medium", "high", "critical"]

    Critical: Production Sensitive data or function - PII data, Financial or other sensitive data
    High: Production/Non-production Sensitive data or function -data not in sensitive category-
    but high importance to business
    Medium: Non-Production  - Criticality to business
    Low: Non-production - tests

    Essential: The vulnerable component directly provides capabilities that constitute at least one MEF
    for at least one entity;
    """

    query = {"Environment": environment, "AssetType": asset_type, "AssetCriticality": asset_criticality}
    vector = calculate_impact(query)
    return vector
